# Rest-a-culous

An opinionated ExpressJs based REST API generator.

## Installation
To use the generator simply install it globally using NPM:

`npm install -g restaculous`

## Features
The generator produces a fully functional application that can be used as a starting point for build a RESTful API based
on [NodeJs](https://nodejs.org), [ExpressJs](http://expressjs.com) and [MongoDB](https://www.mongodb.com/).

The application generated has the following features:

  - [x] Fully tested API endpoints with **CRUD** operations
  - [x] Automatically generate API documentation
  - [ ] Token based **Authentication** and **Authorization**
  - [ ] Linting of source to ensure proper code formatting

## Structure
The generate application structure looks as follows.
```
..
.
 |--config/                 # configuration settings
 |--controllers/            # controllers based on models
 |--dal/                    # data access layer containing abstractions over models
 |--doc/                    # autogenerated documentation for API endpoints
 |--lib/                    # utility library to do common tasks
 |--models/                 # the underlying models of the system
 |--routes/                 # REST-based API endpoints
 |--test/                   # tests for the entire code base
 |--.gitignore              # common file and folders to ignore by git
 |--app.js                  # the applications main entry point
 |--package.json            # specifies modules/packages used in the app
 |--README.md               # an introductory text about the application
```

## Generate from a `settings.json` file
You can generate an entire application by simply supplying a **json** settings file in the following format.

```json
{
  "name": "[name of the application]",
  "description": "[small description about the application]",
  "author": "[author name] <author email>",
  "directory": "[the directory where to output the new application]",
  "repository": {
    "type": "[repository type]",
    "url": "[repository address]"
  },
  "config": [
    {
    "name": "[name of the configuration]",
    "value":"[value of the configuration]",
    "comment": "[short comment about the config]"
    }
  ],
  "models": [
    {
      "name": "[model name(capitalized)]",
      "attributes": [
        {
          "name": "[name of the attribute]",
          "type": "[the data type of the attribute]",
          "desc": "[a small description about the attribute]",
          "example": "[an example of the attribute]",
          "isPrivate": "[indicates the attribute will not be returned, e.g. password (optional)]",
          "isAuto": "[indicates that the value for the attribute will be generated automatically, e.g. date_created (optional)]"
        }
      ],
      "relations": [{
        "name": "[name of a related model(capitalized)]",
        "referenceType": "multiple"
      }]
    }
  ]
}
```

Here is an example `settings.json` file:

```json
{
  "name": "Movies",
  "description": "A REST API endpoint for a Movie streaming app.",
  "author": "Tsegaselassie Tadesse <tsegaselassi@gmail.com>",
  "directory": "/Users/tsega/Projects/movies-api-tmp",
  "repository": {
    "type": "git",
    "url": "http://git.gebeya.com/tsega/movie-api.git"
  },
  "config": [
    {"name": "HTTP_PORT", "value":"process.env.HTTP_PORT || 8000", "comment": "HTTP PORT"},
    {"name": "MONGODB_URL", "value":"'mongodb://localhost/movies'", "comment": "Mongodb URL"},
    {"name": "SALT_LENGTH", "value":"13", "comment": "SALT VALUE LENGTH"},
    {"name": "TOKEN_LENGTH", "value":"253", "comment": "TOKEN LENGTH"},
    {"name": "MAX_PAGE_SIZE", "value":"100", "comment": "DEFAULT PAGE SIZE"},
    {"name": "DEFAULT_SORT", "value":"'last_updated'", "comment": "DEFAULT SORT FIELD"}
  ],
  "models": [
    {
      "name": "Movie",
      "attributes": [
        {
          "name": "title",
          "type": "String",
          "desc": "The title of the movie.",
          "example": "Forest Gump"
        },
        {
          "name": "genre",
          "type": "String",
          "desc": "The category of the movie.",
          "example": "Action",
          "isPrivate": true
        }
      ],
      "relations": [{
        "name": "Actor",
        "referenceType": "multiple"
      }]
    },
    {
      "name": "Actor",
      "attributes": [
        {
          "name": "firstName",
          "type": "String",
          "desc": "The first name of the actor.",
          "example": "Tom"
        },
        {
          "name": "lastName",
          "type": "String",
          "desc": "The last name of the actor.",
          "example": "Hanks"
        }
      ],
      "relations": [{
        "name": "Movie",
        "referenceType": "multiple"
      }]
    },
    {
      "name": "Playlist",
      "attributes": [
        {
          "name": "name",
          "type": "String",
          "desc": "The name of the playlist.",
          "example": "Classic Movies"
        }
      ],
      "relations": [{
        "name": "Movie",
        "referenceType": "multiple"
      }]
    }
  ]
}
```

> **Notes:**
 - Timestamp attributes, i.e. `createdAt` and `updatedAt` are automatically added to all models.
 - All settings fields in the above example are required except the ones identified as being optional. All **configuration entries** are mandatory; a starter `settings.json` file is given below:


**DON'T FORGET TO REPLACE [app-name]**

```json
{
  "name": "[app-name]",
  "description": "",
  "author": "",
  "directory": "",
  "repository": {
    "type": "",
    "url": ""
  },
  "config": [
    {"name": "HTTP_PORT", "value":"process.env.HTTP_PORT || 8000", "comment": "HTTP PORT"},
    {"name": "MONGODB_URL", "value":"'mongodb://localhost/[app-name]'", "comment": "Mongodb URL"},
    {"name": "SALT_LENGTH", "value":"13", "comment": "SALT VALUE LENGTH"},
    {"name": "TOKEN_LENGTH", "value":"253", "comment": "TOKEN LENGTH"},
    {"name": "MAX_PAGE_SIZE", "value":"100", "comment": "DEFAULT PAGE SIZE"},
    {"name": "DEFAULT_SORT", "value":"'last_updated'", "comment": "DEFAULT SORT FIELD"}
  ],
  "models": [
    {
      "name": "",
      "attributes": [
        {
          "name": "",
          "type": "",
          "desc": "",
          "example": ""
        }
      ]
    }
  ]
}
```

